@@ -10,6 +10,7 @@ from chainlib.eth.tx import (
 from chainlib.hash import keccak256_string_to_hex
 from chainlib.eth.contract import (
         ABIContractEncoder,
+        ABIContractDecoder,
         ABIContractType,
         abi_decode_single,
         )
@@ -20,13 +21,38 @@ from hexathon import (
         add_0x,
         strip_0x,
         )
+from dexif import from_fixed
 
 # local imports
 from erc20_demurrage_token.data import data_dir
+from erc20_demurrage_token.seal import SealedContract
+from erc20_demurrage_token.expiry import ExpiryContract
 
 logg = logging.getLogger(__name__)
 
 
+class DemurrageRedistribution:
+    
+    def __init__(self, v):
+        d = ABIContractDecoder()
+        v = strip_0x(v)
+        d.typ(ABIContractType.UINT256)
+        d.typ(ABIContractType.UINT256)
+        d.typ(ABIContractType.BYTES32)
+        d.val(v[:64])
+        d.val(v[64:128])
+        d.val(v[128:192])
+        r = d.decode()
+
+        self.period = r[0]
+        self.value = r[1]
+        self.demurrage = from_fixed(r[2])
+
+
+    def __str__(self):
+        return 'period {} value {} demurrage {}'.format(self.period, self.value, self.demurrage)
+
+
 class DemurrageTokenSettings:
 
     def __init__(self):
@@ -38,21 +64,22 @@ class DemurrageTokenSettings:
         self.sink_address = None
 
 
-class DemurrageToken(ERC20):
+    def __str__(self):
+        return 'name {} demurrage level {} period minutes {} sink address {}'.format(
+                self.name,
+                self.demurrage_level,
+                self.period_minutes,
+                self.sink_address,
+                )
+
+
+class DemurrageToken(ERC20, SealedContract, ExpiryContract):
 
     __abi = {}
     __bytecode = {}
-    valid_modes = [
-                'MultiNocap',
-                'SingleNocap',
-                'MultiCap',
-                'SingleCap',
-                ]
 
-    def constructor(self, sender_address, settings, redistribute=True, cap=0, tx_format=TxFormat.JSONRPC):
-        if int(cap) < 0:
-            raise ValueError('cap must be 0 or positive integer')
-        code = DemurrageToken.bytecode(multi=redistribute, cap=cap>0)
+    def constructor(self, sender_address, settings, tx_format=TxFormat.JSONRPC):
+        code = DemurrageToken.bytecode()
         enc = ABIContractEncoder()
         enc.string(settings.name)
         enc.string(settings.symbol)
@@ -60,8 +87,6 @@ class DemurrageToken(ERC20):
         enc.uint256(settings.demurrage_level)
         enc.uint256(settings.period_minutes)
         enc.address(settings.sink_address)
-        if cap > 0:
-            enc.uint256(cap)
         code += enc.get()
         tx = self.template(sender_address, None, use_nonce=True)
         tx = self.set_code(tx, code)
@@ -74,23 +99,8 @@ class DemurrageToken(ERC20):
 
 
     @staticmethod
-    def __to_contract_name(multi, cap):
-        name = 'DemurrageToken'
-        if multi:
-            name += 'Multi'
-        else:
-            name += 'Single'
-        if cap:
-            name += 'Cap'
-        else:
-            name += 'Nocap'
-        logg.debug('bytecode name {}'.format(name))
-        return name
-
-
-    @staticmethod
-    def abi(multi=True, cap=False):
-        name = DemurrageToken.__to_contract_name(multi, cap)
+    def abi(multi=True):
+        name = 'DemurrageTokenSingleNocap'
         if DemurrageToken.__abi.get(name) == None:
             f = open(os.path.join(data_dir, name + '.json'), 'r')
             DemurrageToken.__abi[name] = json.load(f)
@@ -99,8 +109,8 @@ class DemurrageToken(ERC20):
 
 
     @staticmethod
-    def bytecode(multi=True, cap=False):
-        name = DemurrageToken.__to_contract_name(multi, cap)
+    def bytecode(multi=True):
+        name = 'DemurrageTokenSingleNocap'
         if DemurrageToken.__bytecode.get(name) == None:
             f = open(os.path.join(data_dir, name + '.bin'), 'r')
             DemurrageToken.__bytecode[name] = f.read()
@@ -108,6 +118,34 @@ class DemurrageToken(ERC20):
         return DemurrageToken.__bytecode[name]
 
 
+    def increase_allowance(self, contract_address, sender_address, address, value, tx_format=TxFormat.JSONRPC):
+        enc = ABIContractEncoder()
+        enc.method('increaseAllowance')
+        enc.typ(ABIContractType.ADDRESS)
+        enc.typ(ABIContractType.UINT256)
+        enc.address(address)
+        enc.uint256(value)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
+
+
+    def decrease_allowance(self, contract_address, sender_address, address, value, tx_format=TxFormat.JSONRPC):
+        enc = ABIContractEncoder()
+        enc.method('decreaseAllowance')
+        enc.typ(ABIContractType.ADDRESS)
+        enc.typ(ABIContractType.UINT256)
+        enc.address(address)
+        enc.uint256(value)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
+
+
     def add_minter(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
         enc = ABIContractEncoder()
         enc.method('addMinter')
@@ -120,6 +158,17 @@ class DemurrageToken(ERC20):
         return tx
 
 
+    def set_max_supply(self, contract_address, sender_address, cap, tx_format=TxFormat.JSONRPC):
+        enc = ABIContractEncoder()
+        enc.method('setMaxSupply')
+        enc.typ(ABIContractType.UINT256)
+        enc.uint256(cap)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
+
     def remove_minter(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
         enc = ABIContractEncoder()
         enc.method('removeMinter')
@@ -146,6 +195,33 @@ class DemurrageToken(ERC20):
         return tx
 
 
+    def burn(self, contract_address, sender_address, value, tx_format=TxFormat.JSONRPC):
+        enc = ABIContractEncoder()
+        enc.method('burn')
+        enc.typ(ABIContractType.UINT256)
+        enc.uint256(value)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
+
+
+    def total_burned(self, contract_address, sender_address=ZERO_ADDRESS, id_generator=None):
+        j = JSONRPCRequest(id_generator)
+        o = j.template()
+        o['method'] = 'eth_call'
+        enc = ABIContractEncoder()
+        enc.method('totalBurned')
+        data = add_0x(enc.get())
+        tx = self.template(sender_address, contract_address)
+        tx = self.set_code(tx, data)
+        o['params'].append(self.normalize(tx))
+        o['params'].append('latest')
+        o = j.finalize(o)
+        return o
+
+
     def to_base_amount(self, contract_address, value, sender_address=ZERO_ADDRESS, id_generator=None):
         j = JSONRPCRequest(id_generator)
         o = j.template()
@@ -216,18 +292,18 @@ class DemurrageToken(ERC20):
         return o
 
 
-    def to_redistribution(self, contract_address, participants, demurrage_modifier_ppm, value, period, sender_address=ZERO_ADDRESS, id_generator=None):
+    def to_redistribution(self, contract_address, participants, demurrage_modifier, value, period, sender_address=ZERO_ADDRESS, id_generator=None):
         j = JSONRPCRequest(id_generator)
         o = j.template()
         o['method'] = 'eth_call'
         enc = ABIContractEncoder()
         enc.method('toRedistribution')
         enc.typ(ABIContractType.UINT256)
-        enc.typ(ABIContractType.UINT256)
+        enc.typ_literal('int128')
         enc.typ(ABIContractType.UINT256)
         enc.typ(ABIContractType.UINT256)
         enc.uint256(participants)
-        enc.uint256(demurrage_modifier_ppm)
+        enc.uint256(demurrage_modifier)
         enc.uint256(value)
         enc.uint256(period)
         data = add_0x(enc.get())
@@ -246,8 +322,11 @@ class DemurrageToken(ERC20):
         o['method'] = 'eth_call'
         enc = ABIContractEncoder()
         enc.method('toRedistributionPeriod')
-        enc.typ(ABIContractType.BYTES32)
-        enc.bytes32(redistribution)
+        v = strip_0x(redistribution)
+        enc.typ_literal('(uint32,uint72,uint64)')
+        enc.bytes32(v[:64])
+        enc.bytes32(v[64:128])
+        enc.bytes32(v[128:192])
         data = add_0x(enc.get())
         tx = self.template(sender_address, contract_address)
         tx = self.set_code(tx, data)
@@ -257,22 +336,26 @@ class DemurrageToken(ERC20):
         return o
 
 
-    def to_redistribution_participants(self, contract_address, redistribution, sender_address=ZERO_ADDRESS, id_generator=None):
-        j = JSONRPCRequest(id_generator)
-        o = j.template()
-        o['method'] = 'eth_call'
-        enc = ABIContractEncoder()
-        enc.method('toRedistributionParticipants')
-        enc.typ(ABIContractType.BYTES32)
-        enc.bytes32(redistribution)
-        data = add_0x(enc.get())
-        tx = self.template(sender_address, contract_address)
-        tx = self.set_code(tx, data)
-        o['params'].append(self.normalize(tx))
-        o['params'].append('latest')
-        o = j.finalize(o)
-        return o
-
+#    def to_redistribution_participants(self, contract_address, redistribution, sender_address=ZERO_ADDRESS, id_generator=None):
+#        j = JSONRPCRequest(id_generator)
+#        o = j.template()
+#        o['method'] = 'eth_call'
+#        enc = ABIContractEncoder()
+#        enc.method('toRedistributionParticipants')
+#        v = strip_0x(redistribution)
+#        enc.typ_literal('(uint32,uint72,uint104)')
+#        #enc.typ(ABIContractType.BYTES32)
+#        enc.bytes32(v[:64])
+#        enc.bytes32(v[64:128])
+#        enc.bytes32(v[128:192])
+#        data = add_0x(enc.get())
+#        tx = self.template(sender_address, contract_address)
+#        tx = self.set_code(tx, data)
+#        o['params'].append(self.normalize(tx))
+#        o['params'].append('latest')
+#        o = j.finalize(o)
+#        return o
+#
 
     def to_redistribution_supply(self, contract_address, redistribution, sender_address=ZERO_ADDRESS, id_generator=None):
         j = JSONRPCRequest(id_generator)
@@ -280,8 +363,11 @@ class DemurrageToken(ERC20):
         o['method'] = 'eth_call'
         enc = ABIContractEncoder()
         enc.method('toRedistributionSupply')
-        enc.typ(ABIContractType.BYTES32)
-        enc.bytes32(redistribution)
+        v = strip_0x(redistribution)
+        enc.typ_literal('(uint32,uint72,uint64)')
+        enc.bytes32(v[:64])
+        enc.bytes32(v[64:128])
+        enc.bytes32(v[128:192])
         data = add_0x(enc.get())
         tx = self.template(sender_address, contract_address)
         tx = self.set_code(tx, data)
@@ -297,8 +383,11 @@ class DemurrageToken(ERC20):
         o['method'] = 'eth_call'
         enc = ABIContractEncoder()
         enc.method('toRedistributionDemurrageModifier')
-        enc.typ(ABIContractType.BYTES32)
-        enc.bytes32(redistribution)
+        v = strip_0x(redistribution)
+        enc.typ_literal('(uint32,uint72,uint64)')
+        enc.bytes32(v[:64])
+        enc.bytes32(v[64:128])
+        enc.bytes32(v[128:192])
         data = add_0x(enc.get())
         tx = self.template(sender_address, contract_address)
         tx = self.set_code(tx, data)
@@ -325,8 +414,31 @@ class DemurrageToken(ERC20):
         return o
 
 
-    def apply_demurrage(self, contract_address, sender_address):
-        return self.transact_noarg('applyDemurrage', contract_address, sender_address)
+    def set_sink_address(self, contract_address, sender_address, address, tx_format=TxFormat.JSONRPC):
+        enc = ABIContractEncoder()
+        enc.method('setSinkAddress')
+        enc.typ(ABIContractType.ADDRESS)
+        enc.address(address)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
+
+
+    def apply_demurrage(self, contract_address, sender_address, limit=0, tx_format=TxFormat.JSONRPC):
+        if limit == 0:
+            return self.transact_noarg('applyDemurrage', contract_address, sender_address)
+
+        enc = ABIContractEncoder()
+        enc.method('applyDemurrageLimited')
+        enc.typ(ABIContractType.UINT256)
+        enc.uint256(limit)
+        data = enc.get()
+        tx = self.template(sender_address, contract_address, use_nonce=True)
+        tx = self.set_code(tx, data)
+        tx = self.finalize(tx, tx_format)
+        return tx
 
 
     def change_period(self, contract_address, sender_address):
@@ -345,6 +457,8 @@ class DemurrageToken(ERC20):
         return tx
 
 
+
+
     def tax_level(self, contract_address, sender_address=ZERO_ADDRESS):
         return self.call_noarg('taxLevel', contract_address, sender_address=sender_address)
 
@@ -369,29 +483,33 @@ class DemurrageToken(ERC20):
         return self.call_noarg('demurrageAmount', contract_address, sender_address=sender_address)
 
 
-    def supply_cap(self, contract_address, sender_address=ZERO_ADDRESS):
-        return self.call_noarg('supplyCap', contract_address, sender_address=sender_address)
+    def demurrage_timestamp(self, contract_address, sender_address=ZERO_ADDRESS):
+        return self.call_noarg('demurrageTimestamp', contract_address, sender_address=sender_address)
 
 
-    def grow_by(self, contract_address, value, period, sender_address=ZERO_ADDRESS, id_generator=None):
-        j = JSONRPCRequest(id_generator)
-        o = j.template()
-        o['method'] = 'eth_call'
-        enc = ABIContractEncoder()
-        enc.method('growBy')
-        enc.typ(ABIContractType.UINT256)
-        enc.typ(ABIContractType.UINT256)
-        enc.uint256(value)
-        enc.uint256(period)
-        data = add_0x(enc.get())
-        tx = self.template(sender_address, contract_address)
-        tx = self.set_code(tx, data)
-        o['params'].append(self.normalize(tx))
-        o['params'].append('latest')
-        o = j.finalize(o)
-        return o
+    def max_supply(self, contract_address, sender_address=ZERO_ADDRESS):
+        return self.call_noarg('maxSupply', contract_address, sender_address=sender_address)
 
 
+#    def grow_by(self, contract_address, value, period, sender_address=ZERO_ADDRESS, id_generator=None):
+#        j = JSONRPCRequest(id_generator)
+#        o = j.template()
+#        o['method'] = 'eth_call'
+#        enc = ABIContractEncoder()
+#        enc.method('growBy')
+#        enc.typ(ABIContractType.UINT256)
+#        enc.typ(ABIContractType.UINT256)
+#        enc.uint256(value)
+#        enc.uint256(period)
+#        data = add_0x(enc.get())
+#        tx = self.template(sender_address, contract_address)
+#        tx = self.set_code(tx, data)
+#        o['params'].append(self.normalize(tx))
+#        o['params'].append('latest')
+#        o = j.finalize(o)
+#        return o
+#
+
     def decay_by(self, contract_address, value, period, sender_address=ZERO_ADDRESS, id_generator=None):
         j = JSONRPCRequest(id_generator)
         o = j.template()
@@ -418,7 +536,7 @@ class DemurrageToken(ERC20):
         enc = ABIContractEncoder()
         enc.method('getDistribution')
         enc.typ(ABIContractType.UINT256)
-        enc.typ(ABIContractType.UINT256)
+        enc.typ_literal('int128')
         enc.uint256(supply)
         enc.uint256(demurrage_amount)
         data = add_0x(enc.get())
@@ -436,8 +554,11 @@ class DemurrageToken(ERC20):
         o['method'] = 'eth_call'
         enc = ABIContractEncoder()
         enc.method('getDistributionFromRedistribution')
-        enc.typ(ABIContractType.BYTES32)
-        enc.bytes32(redistribution)
+        v = strip_0x(redistribution)
+        enc.typ_literal('(uint32,uint72,uint64)')
+        enc.bytes32(v[:64])
+        enc.bytes32(v[64:128])
+        enc.bytes32(v[128:192])
         data = add_0x(enc.get())
         tx = self.template(sender_address, contract_address)
         tx = self.set_code(tx, data)
@@ -465,7 +586,8 @@ class DemurrageToken(ERC20):
 
     @classmethod
     def parse_demurrage_amount(self, v):
-        return abi_decode_single(ABIContractType.UINT256, v)
+        #        return abi_decode_single(ABIContractType.UINT256, v)
+        return from_fixed(v)
 
 
     @classmethod
@@ -480,8 +602,8 @@ class DemurrageToken(ERC20):
     
     @classmethod
     def parse_redistributions(self, v):
-        return abi_decode_single(ABIContractType.BYTES32, v)
-
+        return strip_0x(v)
+        
 
     @classmethod
     def parse_account_period(self, v):
@@ -502,6 +624,7 @@ class DemurrageToken(ERC20):
     def parse_supply_cap(self, v):
         return abi_decode_single(ABIContractType.UINT256, v)
 
+
     @classmethod
     def parse_grow_by(self, v):
         return abi_decode_single(ABIContractType.UINT256, v)
@@ -525,3 +648,9 @@ class DemurrageToken(ERC20):
     @classmethod
     def parse_resolution_factor(self, v):
         return abi_decode_single(ABIContractType.UINT256, v)
+
+
+    @classmethod
+    def parse_total_burned(self, v):
+        return abi_decode_single(ABIContractType.UINT256, v)
+
